from skimage import io, data
import numpy as np
from skimage import img_as_float as imfloat
from copy import copy

# TODO: see this for sub-classing np arrays to add metadata
# https://stackoverflow.com/questions/34967273/add-metadata-comment-to-numpy-ndarray
# TODO: unit-testing


def extract(image, xRng=[0, 1], yRng=[0, 1], kapa=1.0):

    # Lower- and upper bounds for clipping the x-range
    xLB = xRng[0]
    xUB = xRng[1]

    # Lower- and upper bounds for clipping the yx-range
    yLB = yRng[0]
    yUB = yRng[1]
    
    # Clip the raw image into the region of interest
    imgClip = copy(image)

    # Find and count the pixels with maximum luminousity
    position = np.where(imgClip >= kapa)
    featureCount = position[1].size

    if featureCount > 0:
        for pair_0, pair_1 in zip(position[0], position[1]):
            if pair_0 > 3 and pair_1 > 3:
                imgClip[pair_0-4:pair_0+4, pair_1-4:pair_1+4] = 0

    return featureCount, position, imgClip


if __name__ == "__main__":

    from init import examine
    from matplotlib import pyplot as plt
    
    image = './Orion_1.jpg'
    img_as_array, x_range, y_range = examine(image, verbose=False, graphics=False)
    imgGray = img_as_array @ [0.2126, 0.7152, 0.0722]  # image in grayscale

    fig1, (left, right) = plt.subplots(1, 2)
    left.imshow(imgGray)
    left.set_title('Raw image', fontsize=14)

    print(f'The input image has x_range = {x_range}, and y_range = {y_range}')

    numFeatures, indices, markers = extract(imgGray, xRng=[2110, 2130], yRng=[850, 870], kapa=0.9)

    imgGray[2110:2130, 850:870] = markers[:, :]

    right.imshow(imgGray)
    right.set_title('Marked features', fontsize=14)
    plt.show()
    
    print(numFeatures, indices)



from skimage import img_as_float
from skimage.io import imread, imsave
import numpy as np

# TODO: use class method to define an alternative constructor based on a numpy arrays saved on disk

class Image:
    
    def __init__(self, img_path):
        self.img_path = img_path
        self.img_data = imread(img_path)   # raw image data as an integer-valued array

    def print_info(self):
        """Print out the object type, datatype, and the dimensions of the input object.
        Input: data array of the image which can be of integer or floating point type
        """
        np.set_printoptions(precision=2, linewidth=155)        
        print('Image type: ', type(self.img_data))
        print('Image datatype: ', self.img_data.dtype)
        print('Image shape: ', self.img_data.shape)

    def plot(self, frame_title=None):   
        from matplotlib import pyplot as plt

        gridShow = False
        fig, ax1 = plt.subplots(1, 1)

        # TODO: can we expand imagePath and loop over if it has more than 1 path?
        ax1.imshow(self.img_data)
        ax1.set_title(frame_title, fontsize=14)
        ax1.grid(gridShow)
        plt.show()

    def save(self):
        "save the image data array to disk"
        # TODO: Find the installation path; then cd to data/
        # TODO: use hdf5
        np.save('../data/input_image', self.img_data)   

    def size(self):
        "Return the x and y ranges of the input array, which can be of integer or floating point type"
        x_max = self.img_data.shape[1]
        y_max = self.img_data.shape[0]

        return x_max, y_max


class FloatImage(Image):
    def __init__(self, img_path):
        Image.__init__(self, img_path)
        self.img_data = img_as_float(imread(self.img_path))   # image data as a floating point array

        
class GrayImage(FloatImage):
    def __init__(self, img_path):
        FloatImage.__init__(self, img_path)
        self.img_data = self.img_data @ [0.2126, 0.7152, 0.0722]  # image in gray-scale form

    def plot(self, frame_title=None):   
        from matplotlib import pyplot as plt

        gridShow = False
        fig, ax1 = plt.subplots(1, 1)

        # TODO: can we expand imagePath and loop over if it has more than 1 path?
        ax1.imshow(self.img_data, cmap='gray')
        ax1.set_title(frame_title, fontsize=14)
        ax1.grid(gridShow)
        plt.show()
        
    def extract(self, xRng=[0, 1], yRng=[0, 1], kapa=1.0):
        from copy import copy
        # Lower- and upper bounds for clipping the x-range
        xLB = xRng[0]
        xUB = xRng[1]

        # Lower- and upper bounds for clipping the yx-range
        yLB = yRng[0]
        yUB = yRng[1]
    
        # Clip the raw image into the region of interest
        imgClip = copy(self.img_data)

        # Find and count the pixels with maximum luminousity
        position = np.where(imgClip >= kapa)
        featureCount = position[1].size

        if featureCount > 0:
            for pair_0, pair_1 in zip(position[0], position[1]):
                if pair_0 > 3 and pair_1 > 3:
                    imgClip[pair_0-4:pair_0+4, pair_1-4:pair_1+4] = 0

        return featureCount, position, imgClip


    
image_path = '../tests/beam_0.jpg'
img = Image(image_path)

print(img.img_data[0:5, 0, 0])

img.print_info()

###img.plot(frame_title='Raw image')

x_range, y_range = img.size()
print(x_range, y_range)


imgFloat = FloatImage(image_path)
print(imgFloat.img_data[0:5, 0, 0])

imgFloat.print_info()

###imgFloat.plot(frame_title='Float image')

x_range, y_range = imgFloat.size()
print(x_range, y_range)


imgGray = GrayImage(image_path)
print(imgGray.img_data[0:5, 0])

imgGray.print_info()

###imgGray.plot(frame_title='Gray image')

x_range, y_range = imgGray.size()
print(x_range, y_range)

numFeatures, indices, markers = imgGray.extract(xRng=[0, x_range], yRng=[0, y_range],
                                                kapa=0.9998)

print(f'''
        Total number of features detected: {numFeatures}. If this is too many, consider increasing  
        the threshold input.
    ''')
#print(img.__dict__)



print('Done')
        



from skimage import io, data
import numpy as np
from matplotlib import pyplot as plt
from skimage import img_as_float as imfloat
import os


def examine(imagePath, save=False, verbose=False, graphics=False):
    image = imfloat(io.imread(imagePath))

    if verbose:
        np.set_printoptions(precision=2, linewidth=155)
        
        print('Image type: ', type(image))
        print('Image datatype: ', image.dtype)
        print('Image shape: ', image.shape)

    if graphics:
        gridShow = False
        fig, ax1 = plt.subplots(1, 1)

        # TODO: can we expand imagePath and loop over if it has more than 1 path?
        ax1.imshow(image, cmap='gray')
        ax1.set_title('Input image', fontsize=14)
        ax1.grid(gridShow)
        plt.show()

    # save the image
    if save:
        work_dir = os.getcwd()   # Current working directory
        np.save(f'{work_dir}/data/input_image', image)   # TODO:data/ should be created by init.setup() needs test?

    # return the x, and y ranges
    x_max = image.shape[0]
    y_max = image.shape[1]

    return image, x_max, y_max


def setup():
    # prepare the data directory structure for the program
    try:
        #TODO: for more clarity, name the base dir. estare_data instead of just data
        os.mkdir('./data')
        os.mkdir('./data/features')
        os.mkdir('./data/refuse')
        
        os.mkdir('./data/features/coordinates')
        os.mkdir('./data/features/pixels')

        os.mkdir('./data/refuse/coordinates')        
        os.mkdir('./data/refuse/pixels')                
    except OSError:
        print('Failed to create data directories. This may be due to the lack of write permission.')
            

if __name__ == '__main__':

    #TODO:++++++++++++ This test is broken; ./tests/test.jpg does NOT exist ++++++++++++++++++++++
    xs
    image, x_range, y_range = examine('./tests/test.jpg', save=False, verbose=True, graphics=False)
    print(f'Image has x-range = {x_range}, and y-range = {y_range}')
    
    imgGray = image @ [0.2126, 0.7152, 0.0722]  # image in grayscale

    gridShow = False
    fig, ax1 = plt.subplots(1, 1)

    ax1.imshow(imgGray, cmap='gray')
    ax1.set_title('Test image', fontsize=14)
    ax1.grid(gridShow)
    plt.show()



import numpy as np
from estare.src.init import examine, setup
from estare.src.feature import extract
from estare.src.align import align
    
#from src.init import examine, setup
#from src.feature import extract
#from src.align import align

from matplotlib import pyplot as plt
import argparse
import os
from skimage.io import imsave

def main():
    
    parser = argparse.ArgumentParser(prog='estare', description='''Program for aligning and stacking astro photos.''',
                                     epilog='''estare is a Persian word for star.''')

    sub_parsers = parser.add_subparsers()

    # create the parser for the feature detection command
    parse_feature = sub_parsers.add_parser('scan', help='''Scan an input image to find suitable features such as (a) bright 
    star(s) that can be used for aligning multiple frames.''')

    parse_feature.add_argument('image', action='store', type=str, help='The input image to be scanned.')
    parse_feature.add_argument('kapa', action='store', type=float, help='''Threshold for picking pixels of a certain  
    brightness. It takes a floating point values from the range [0, 1]. Pick a larger value to
    limit the detected features to those of a higher brightness.''')
    parse_feature.set_defaults(Mode='scan')

    parse_stack = sub_parsers.add_parser('stack', help='''This mode should normally be used after the alignment features
    have been detected in the scan mode. The stack mode uses the coordinates of the 
    features to align two images. It then stacks them up. Two input images along 
    with one identical feature for each image are required.''')

    parse_stack.add_argument('layer_1', help='Path and name of the reference image (base layer)')
    parse_stack.add_argument('layer_2', help='Path and name of the second image (top layer)')
    parse_stack.add_argument('feature_1', help='''Path and name of a .npy file containing the coordinates of the feature in 
    the reference layer''')
    parse_stack.add_argument('feature_2', help='''Path and name of a .npy file containing the coordinates of the feature in
    the second layer which will be lined up with the identical feature in the reference 
    layer.''')
    parse_stack.set_defaults(Mode='stack')

    args = parser.parse_args()

    # Perform the initial setup
    setup()
    
    if args.Mode == 'scan':
        
        # handle input arguments
        imagePath = args.image
        threshold = args.kapa

        img_as_array, x_range, y_range = examine(imagePath, save=True, verbose=False, graphics=False)
        # Replace with:
        # img_as_array = FloatImage(imagePath)
        # img_as_array.print_info()
        # img_as_array.save()
        # x_range, y_range = img_as_array.size()
    
        print(f'Image has x-range = {x_range}, and y-range = {y_range}')

        imgGray = img_as_array @ [0.2126, 0.7152, 0.0722]  # image in grayscale
        # Replace with (makes FloatImage redundant):
        # imgGray = GrayImage(imagePath)
        # imgGray.print_info()
        # imgGray.save()
        # x_range, y_range = imgGray.size()

        fig2, frame = plt.subplots(1, 1)
        image_features = frame.imshow(imgGray, cmap='gray')
        frame.set_title('Marked features', fontsize=14)

        numFeatures, indices, markers = extract(imgGray, xRng=[0, x_range], yRng=[0, y_range],
                                                kapa=threshold)
        # Replace with:
        # numFeatures, indices, markers = imgGray.extract(xRng=[0, x_range], yRng=[0, y_range], kapa=threshold)

        print(f'''
        Total number of features detected: {numFeatures}. If this is too many, consider increasing  
        the threshold input.
        ''')

        # Loop over features and ask the user if a feature is of interest to be saved
        keepFeature = False

        xy_FeatureCount = 0   # number of already-saved feature coordinates initialized to zero
        binFeatureCount = 0   # number of already-saves feature images in binary format initialized to zero
        imgFeatureCount = 0   # number of already-saves feature images in .png format initialized to zero

        xy_RefusedCount = 0   # number of already-saved refused coordinates initialized to zero
        binRefusedCount = 0   # number of already-saves refused images in binary format initialized to zero
        imgRefusedCount = 0   # number of already-saves refused images in .png format initialized to zero  

        for files in os.listdir('./data/features'):
            if files.endswith('.png'):
                imgFeatureCount += 1

        for files in os.listdir('./data/features/pixels'):
            if files.endswith('.npy'):
                binFeatureCount += 1

        for files in os.listdir('./data/features/coordinates'):
            if files.endswith('.npy'):
                xy_FeatureCount += 1


        for files in os.listdir('./data/refuse'):
            if files.endswith('.png'):
                imgRefusedCount += 1

        for files in os.listdir('./data/refuse/pixels'):
            if files.endswith('.npy'):
                binRefusedCount += 1

        for files in os.listdir('./data/refuse/coordinates'):
            if files.endswith('.npy'):
                xy_RefusedCount += 1    
            
        selectedFeatures = 0  # counter
        refusedFeatures = 0  # counter
    
    
        # TODO: add a feature to use matplotlib.ginput
        if numFeatures > 0:
            print('''A detected feature is displayed by a black square. Press any key to save the current feature, or click 
            the mouse to discard it...''')
            for pair_0, pair_1 in zip(indices[0], indices[1]):
                if pair_0 > 3 and pair_1 > 3:
                    arrow = frame.annotate('O', xy=(pair_1, pair_0), arrowprops=dict(color='lime',arrowstyle='->'))
                    plt.draw()  # , plt.pause(0.01)
                    btnpress = plt.waitforbuttonpress(-1)
                    if btnpress:
                        selectedFeatures += 1
                        np.save('./data/features/coordinates/feature_{}'.format(xy_FeatureCount), [pair_0, pair_1])
                        xy_FeatureCount += 1
                        np.save('./data/features/pixels/feature_{}_pixels'.format(binFeatureCount), imgGray[pair_0-10:pair_0+10, pair_1-10:pair_1+10])
                        binFeatureCount += 1
                        imsave('./data/features/feature_{}_pixels.png'.format(imgFeatureCount), imgGray[pair_0-10:pair_0+10, pair_1-10:pair_1+10])
                        imgFeatureCount += 1
                        # TODO: Find install path and cd to data
                        plt.waitforbuttonpress(0.1)
                    else:
                        refusedFeatures += 1
                        np.save('./data/refuse/coordinates/refused_{}'.format(xy_RefusedCount), [pair_0, pair_1])
                        xy_RefusedCount += 1
                        np.save('./data/refuse/pixels/refused_{}_pixels'.format(binRefusedCount), imgGray[pair_0-10:pair_0+10, pair_1-10:pair_1+10])
                        binRefusedCount += 1
                        imsave('./data/refuse/refused_{}_pixels.png'.format(imgRefusedCount), imgGray[pair_0-10:pair_0+10, pair_1-10:pair_1+10])
                        imgRefusedCount += 1
                    arrow.remove()   # removing the arrow must be the last thing to do at the end of the if-block
                
        print('Feature detection completed.')
        plt.show()

    elif args.Mode == 'stack':

        # unpack the arguments
        img_1 = args.layer_1
        img_2 = args.layer_2
        pivot_1 = args.feature_1
        pivot_2 = args.feature_2

        # align and stack the two input frames
        stacked_frame = align(img_1, img_2, pivot_1, pivot_2, save=True)


if __name__ == '__main__':
    main()



import numpy as np



def arc(position_vector, origo=np.array([0., 0.]), radians=False):
    
    vector_size = np.size(position_vector)   
    position_matrix = np.zeros( (vector_size//2, vector_size), dtype=float)

    for i in range(vector_size//2):
        col_index = i*2
        position_matrix[i, col_index:col_index+2] = position_vector[col_index], position_vector[col_index+1] 


    sqrd_pos = np.dot(position_matrix, position_vector)
    return sqrd_pos


if __name__ == '__main__':
    pos_vec = np.array([0, 1, 2, 3, 4, 4, 6, 7])
    pos_mtx = arc(pos_vec)
    print(pos_mtx)



import numpy as np


def reference_markers():

    """Returns two pairs of coordinates: 
          targets: these are two reference points
          markers: these are in fact the targets that have been 
                   shifted and rotated. The aim would be to 
                   reverse the effect of the shift and rotation
                   and bring the marker coordinates back to the 
                   target coordinates.
    """
    
    # targets are the original coordinates
    targets = np.array([[1953, 1270],
                       [2087, 1289]])

    # markers include offsets and rotation 
    markers = np.array([[1969, 1281],
                       [2102, 1301]])

    return targets, markers




import numpy as np

#import calculate


'''
We have:
- a function to calculate angle recieves a reference(default 0, 0) and a feature coordinate
- a function that takes an angle and a pair of coordinates and rotates the point (below)
- a function that takes [del_x, del_y] and a pair of coordinates and translates the point (below)

--------------->  Y
|*
| *
|  *
|   *
|
v

X

'''

# @profile
# def arc(position, origo=np.array([0., 0.]), radians=False):
#     '''
# Returns the angle between the input coordinate and the vertical axis Y
# '''
#     rho = np.sqrt( (position[0]-origo[0])**2 + (position[1]-origo[1])**2 )
#     angle = np.arccos( (position[0]-origo[0]) / rho )
    
#     if radians==False:
#         angle = angle*180.0/np.pi
        
#     return angle

#@profile
def arc(position, origo=np.array([0., 0.]), radians=False):
    """Returns the angle between the input coordinate and the vertical 
       axis Y using the projection principle
    """

    #TODO: position[0,0] or generally position=origin should raise error
    # in fact division by rho = 0 should raise error
    position_norm = np.sqrt((position[0]-origo[0])**2 + (position[1]-origo[1])**2)
    # computing norm via np.linalg.norm(position) is much slower
    if radians==False:
        return position_norm, 180.0 * (np.arccos( (position[0]-origo[0]) / position_norm )) / np.pi
    else:
        return position_norm, np.arccos( (position[0]-origo[0]) / position_norm )


    
# apply rotation
#@profile
def rotate(position, angle, origo=np.array([0., 0.]), radians=False, discrete=True):
    #TODO: position [0,0] cannot be rotated

    rho, theta = arc(position, radians=radians)
    theta += angle
    #theta = angle + arc(position, radians=radians)

    # Use the compiled version of arc():
    # theta = angle + calculate.get_angle(position, radians=radians)

    if radians==False:
        theta = theta*(np.pi/180.0)   

    # rho now comes directly from arc()
    #rho = np.sqrt((position[0]-origo[0])**2 + (position[1]-origo[1])**2)
    
    x = rho*np.cos(theta)
    y = rho*np.sin(theta)

    if discrete:
        rotated_position = np.rint([x, y])     
        return np.array( [int(rotated_position[0]), int(rotated_position[1])] )
    else:
        return np.array( [x, y] )









from estare.src.rotate import rotate


def stack(img_1, img_2, x_range, y_range, del_x, del_y, theta, discrete=True):

    for i in range(x_range):
        for j in range(y_range):
            #uncorrected_coor = [x_array[i, j], y_array[i, j]]
            uncorrected_coor = [i-del_x, j-del_y]
            
            corrected_coor = rotate(uncorrected_coor, -theta, discrete=discrete)
            
            #x_array[i, j] = corrected_coor[0]
            #y_array[i, j] = corrected_coor[1]
            
            x_corrected = corrected_coor[0]
            y_corrected = corrected_coor[1]

            # stack +needs unittesting if the stacking algorithm is to be changed+
            #if x_array[i, j] > 0 and x_array[i, j] < x_range and y_array[i, j] > 0 and y_array[i, j] < y_range:
            #    img_1[x_array[i, j], y_array[i, j]] += img_2[i, j]   # Broadcasting to all three channels is implicit
                
            if x_corrected > 0 and x_corrected < x_range and y_corrected > 0 and y_corrected < y_range:
                img_1[x_corrected, y_corrected] += img_2[i, j]   # Broadcasting to all three channels is implicit

    return img_1
                
            
